package dev.shadowsoffire.apotheosis.adventure;

import dev.shadowsoffire.apotheosis.adventure.socket.gem.GemInstance;
import dev.shadowsoffire.apotheosis.adventure.affix.effect.MagicalArrowAffix;
import dev.shadowsoffire.apotheosis.adventure.affix.effect.FestiveAffix;
import dev.shadowsoffire.apotheosis.adventure.affix.effect.OmneticAffix;
import dev.shadowsoffire.apotheosis.adventure.affix.effect.RadialAffix;
import net.minecraft.world.entity.Mob;
import net.minecraftforge.eventbus.api.Event;
import net.minecraft.world.entity.animal.AbstractGolem;
import net.minecraftforge.event.entity.living.LivingEvent;
import dev.shadowsoffire.placebo.events.GetEnchantmentLevelEvent;
import java.util.List;
import net.minecraft.core.BlockPos;
import net.minecraft.world.level.Level;
import net.minecraft.world.level.ItemLike;
import net.minecraft.world.phys.AABB;
import dev.shadowsoffire.placebo.events.AnvilLandEvent;
import net.minecraft.world.level.ServerLevelAccessor;
import dev.shadowsoffire.apotheosis.adventure.loot.LootRarity;
import dev.shadowsoffire.apotheosis.adventure.loot.LootController;
import net.minecraft.world.entity.MobSpawnType;
import net.minecraftforge.event.entity.living.MobSpawnEvent;
import net.minecraftforge.event.entity.player.PlayerEvent;
import dev.shadowsoffire.apotheosis.adventure.affix.effect.TelepathicAffix;
import net.minecraftforge.event.entity.living.LivingDeathEvent;
import net.minecraft.world.entity.item.ItemEntity;
import dev.shadowsoffire.apotheosis.adventure.socket.gem.GemRegistry;
import net.minecraft.world.entity.player.Player;
import dev.shadowsoffire.apotheosis.adventure.compat.GameStagesCompat;
import dev.shadowsoffire.placebo.reload.WeightedDynamicRegistry;
import net.minecraft.server.level.ServerLevel;
import net.minecraftforge.common.util.FakePlayer;
import net.minecraft.world.entity.monster.Monster;
import net.minecraft.server.level.ServerPlayer;
import net.minecraftforge.event.entity.living.LivingDropsEvent;
import net.minecraftforge.event.level.BlockEvent;
import net.minecraftforge.event.entity.living.ShieldBlockEvent;
import java.util.function.Predicate;
import com.google.common.base.Predicates;
import net.minecraft.world.InteractionResult;
import dev.shadowsoffire.placebo.events.ItemUseEvent;
import java.util.Iterator;
import net.minecraft.world.damagesource.DamageSource;
import dev.shadowsoffire.apotheosis.Apoth;
import net.minecraft.world.entity.projectile.Projectile;
import net.minecraftforge.event.entity.ProjectileImpactEvent;
import net.minecraft.world.entity.Entity;
import net.minecraft.world.entity.LivingEntity;
import net.minecraft.world.entity.projectile.AbstractArrow;
import net.minecraftforge.event.entity.EntityJoinLevelEvent;
import net.minecraftforge.eventbus.api.EventPriority;
import net.minecraft.world.damagesource.DamageTypes;
import net.minecraftforge.event.entity.living.LivingHurtEvent;
import dev.shadowsoffire.apotheosis.adventure.affix.AffixInstance;
import dev.shadowsoffire.apotheosis.adventure.affix.Affix;
import dev.shadowsoffire.placebo.reload.DynamicHolder;
import java.util.Map;
import net.minecraft.world.entity.EquipmentSlot;
import dev.shadowsoffire.apotheosis.adventure.socket.SocketedGems;
import net.minecraft.world.item.ItemStack;
import dev.shadowsoffire.apotheosis.adventure.affix.AffixHelper;
import java.util.function.BiConsumer;
import java.util.Objects;
import dev.shadowsoffire.apotheosis.adventure.loot.LootCategory;
import dev.shadowsoffire.apotheosis.adventure.socket.SocketHelper;
import net.minecraftforge.event.ItemAttributeModifierEvent;
import net.minecraftforge.eventbus.api.SubscribeEvent;
import dev.shadowsoffire.apotheosis.adventure.commands.AffixCommand;
import dev.shadowsoffire.apotheosis.adventure.commands.BossCommand;
import dev.shadowsoffire.apotheosis.adventure.commands.SocketCommand;
import dev.shadowsoffire.apotheosis.adventure.commands.GemCommand;
import dev.shadowsoffire.apotheosis.adventure.commands.ModifierCommand;
import dev.shadowsoffire.apotheosis.adventure.commands.LootifyCommand;
import dev.shadowsoffire.apotheosis.adventure.commands.CategoryCheckCommand;
import dev.shadowsoffire.apotheosis.adventure.commands.RarityCommand;
import dev.shadowsoffire.apotheosis.Apotheosis;
import java.util.concurrent.atomic.AtomicBoolean;

public class AdventureEvents
{
    private static ThreadLocal<AtomicBoolean> reentrantLock;
    
    @SubscribeEvent
    public void cmds(final Apotheosis.ApotheosisCommandEvent e) {
        RarityCommand.register(e.getRoot());
        CategoryCheckCommand.register(e.getRoot());
        LootifyCommand.register(e.getRoot());
        ModifierCommand.register(e.getRoot());
        GemCommand.register(e.getRoot());
        SocketCommand.register(e.getRoot());
        BossCommand.register(e.getRoot());
        AffixCommand.register(e.getRoot());
    }
    
    @SubscribeEvent
    public void affixModifiers(final ItemAttributeModifierEvent e) {
        final ItemStack stack = e.getItemStack();
        if (stack.m_41782_()) {
            final SocketedGems gems = SocketHelper.getGems(stack);
            final LootCategory forItem = LootCategory.forItem(stack);
            final EquipmentSlot slotType = e.getSlotType();
            Objects.requireNonNull(e);
            gems.addModifiers(forItem, slotType, (BiConsumer)e::addModifier);
            final Map<DynamicHolder<? extends Affix>, AffixInstance> affixes = AffixHelper.getAffixes(stack);
            affixes.forEach((afx, inst) -> {
                e.getSlotType();
                Objects.requireNonNull(e);
                final EquipmentSlot equipmentSlot;
                inst.addModifiers(equipmentSlot, (BiConsumer)e::addModifier);
            });
        }
    }
    
    @SubscribeEvent(priority = EventPriority.HIGH)
    public void preventBossSuffocate(final LivingHurtEvent e) {
        if (e.getSource().m_276093_(DamageTypes.f_268612_) && e.getEntity().getPersistentData().m_128441_("apoth.boss")) {
            e.setCanceled(true);
        }
    }
    
    @SubscribeEvent(priority = EventPriority.HIGH)
    public void fireArrow(final EntityJoinLevelEvent e) {
        final Entity entity = e.getEntity();
        if (entity instanceof final AbstractArrow arrow) {
            if (!arrow.getPersistentData().m_128471_("apoth.generated")) {
                final Entity 19749_ = arrow.m_19749_();
                if (19749_ instanceof final LivingEntity user) {
                    ItemStack bow = user.m_21211_();
                    if (bow.m_41619_()) {
                        bow = user.m_21205_();
                        if (bow.m_41619_() || !LootCategory.forItem(bow).isRanged()) {
                            bow = user.m_21206_();
                        }
                    }
                    if (bow.m_41619_()) {
                        return;
                    }
                    SocketHelper.getGems(bow).onArrowFired(user, arrow);
                    AffixHelper.streamAffixes(bow).forEach(a -> a.onArrowFired(user, arrow));
                    AffixHelper.copyFrom(bow, (Entity)arrow);
                }
            }
        }
    }
    
    @SubscribeEvent
    public void impact(final ProjectileImpactEvent e) {
        final Projectile projectile = e.getProjectile();
        if (projectile instanceof final AbstractArrow arrow) {
            SocketHelper.getGemInstances(arrow).forEach(inst -> inst.onArrowImpact(arrow, e.getRayTraceResult()));
            final Map<DynamicHolder<? extends Affix>, AffixInstance> affixes = AffixHelper.getAffixes(arrow);
            affixes.values().forEach(inst -> inst.onArrowImpact(arrow, e.getRayTraceResult(), e.getRayTraceResult().m_6662_()));
        }
    }
    
    @SubscribeEvent(priority = EventPriority.LOW)
    public void onDamage(final LivingHurtEvent e) {
        Apoth.Affixes.MAGICAL.getOptional().ifPresent(afx -> afx.onHurt(e));
        final DamageSource src = e.getSource();
        final LivingEntity ent = e.getEntity();
        float amount = e.getAmount();
        for (final ItemStack s : ent.m_20158_()) {
            amount = SocketHelper.getGems(s).onHurt(src, ent, amount);
            final Map<DynamicHolder<? extends Affix>, AffixInstance> affixes = AffixHelper.getAffixes(s);
            for (final AffixInstance inst : affixes.values()) {
                amount = inst.onHurt(src, ent, amount);
            }
        }
        e.setAmount(amount);
    }
    
    @SubscribeEvent
    public void onItemUse(final ItemUseEvent e) {
        final ItemStack s = e.getItemStack();
        final InteractionResult socketRes = SocketHelper.getGems(s).onItemUse(e.getContext());
        if (socketRes != null) {
            e.setCanceled(true);
            e.setCancellationResult(socketRes);
        }
        final InteractionResult afxRes = AffixHelper.streamAffixes(s).map(afx -> afx.onItemUse(e.getContext())).filter(Predicates.notNull()).findFirst().orElse(null);
        if (afxRes != null) {
            e.setCanceled(true);
            e.setCancellationResult(afxRes);
        }
    }
    
    @SubscribeEvent
    public void shieldBlock(final ShieldBlockEvent e) {
        final ItemStack stack = e.getEntity().m_21211_();
        final Map<DynamicHolder<? extends Affix>, AffixInstance> affixes = AffixHelper.getAffixes(stack);
        float blocked = e.getBlockedDamage();
        blocked = SocketHelper.getGems(stack).onShieldBlock(e.getEntity(), e.getDamageSource(), blocked);
        for (final AffixInstance inst : affixes.values()) {
            blocked = inst.onShieldBlock(e.getEntity(), e.getDamageSource(), blocked);
        }
        if (blocked != e.getOriginalBlockedDamage()) {
            e.setBlockedDamage(blocked);
        }
    }
    
    @SubscribeEvent
    public void blockBreak(final BlockEvent.BreakEvent e) {
        final ItemStack stack = e.getPlayer().m_21205_();
        SocketHelper.getGems(stack).onBlockBreak(e.getPlayer(), e.getLevel(), e.getPos(), e.getState());
        AffixHelper.streamAffixes(stack).forEach(inst -> inst.onBlockBreak(e.getPlayer(), e.getLevel(), e.getPos(), e.getState()));
    }
    
    @SubscribeEvent(priority = EventPriority.HIGH)
    public void dropsHigh(final LivingDropsEvent e) {
        final Entity 7639_ = e.getSource().m_7639_();
        if (7639_ instanceof final ServerPlayer p) {
            if (e.getEntity() instanceof Monster) {
                if (p instanceof FakePlayer) {
                    return;
                }
                final float chance = AdventureConfig.gemDropChance + (e.getEntity().getPersistentData().m_128441_("apoth.boss") ? AdventureConfig.gemBossBonus : 0.0f);
                if (p.f_19796_.m_188501_() <= chance) {
                    final Entity ent = (Entity)e.getEntity();
                    e.getDrops().add(new ItemEntity(ent.m_9236_(), ent.m_20185_(), ent.m_20186_(), ent.m_20189_(), GemRegistry.createRandomGemStack(p.f_19796_, (ServerLevel)p.m_9236_(), p.m_36336_(), new Predicate[] { WeightedDynamicRegistry.IDimensional.matches(p.m_9236_()), GameStagesCompat.IStaged.matches((Player)p) }), 0.0, 0.0, 0.0));
                }
            }
        }
    }
    
    @SubscribeEvent(priority = EventPriority.LOW)
    public void drops(final LivingDropsEvent e) {
        Apoth.Affixes.FESTIVE.getOptional().ifPresent(afx -> afx.drops(e));
    }
    
    @SubscribeEvent(priority = EventPriority.LOW)
    public void deathMark(final LivingDeathEvent e) {
        Apoth.Affixes.FESTIVE.getOptional().ifPresent(afx -> afx.markEquipment(e));
    }
    
    @SubscribeEvent(priority = EventPriority.LOWEST)
    public void dropsLowest(final LivingDropsEvent e) {
        TelepathicAffix.drops(e);
    }
    
    @SubscribeEvent(priority = EventPriority.LOWEST, receiveCanceled = true)
    public void festive_removeMarker(final LivingDropsEvent e) {
        Apoth.Affixes.FESTIVE.getOptional().ifPresent(afx -> afx.removeMarker(e));
    }
    
    @SubscribeEvent
    public void harvest(final PlayerEvent.HarvestCheck e) {
        Apoth.Affixes.OMNETIC.getOptional().ifPresent(afx -> afx.harvest(e));
    }
    
    @SubscribeEvent(priority = EventPriority.HIGHEST)
    public void speed(final PlayerEvent.BreakSpeed e) {
        Apoth.Affixes.OMNETIC.getOptional().ifPresent(afx -> afx.speed(e));
    }
    
    @SubscribeEvent(priority = EventPriority.LOW)
    public void onBreak(final BlockEvent.BreakEvent e) {
        Apoth.Affixes.RADIAL.getOptional().ifPresent(afx -> afx.onBreak(e));
    }
    
    @SubscribeEvent(priority = EventPriority.LOW)
    public void special(final MobSpawnEvent.FinalizeSpawn e) {
        if (e.getSpawnType() == MobSpawnType.NATURAL && e.getLevel().m_213780_().m_188501_() <= AdventureConfig.randomAffixItem && e.getEntity() instanceof Monster) {
            final Player player = e.getLevel().m_45924_(e.getX(), e.getY(), e.getZ(), -1.0, false);
            if (player == null) {
                return;
            }
            final ItemStack affixItem = LootController.createRandomLootItem(e.getLevel().m_213780_(), (LootRarity)null, player, (ServerLevelAccessor)e.getEntity().m_9236_());
            if (affixItem.m_41619_()) {
                return;
            }
            affixItem.m_41784_().m_128379_("apoth_rspawn", true);
            final LootCategory cat = LootCategory.forItem(affixItem);
            final EquipmentSlot slot = cat.getSlots()[0];
            e.getEntity().m_8061_(slot, affixItem);
            e.getEntity().m_21508_(slot);
        }
    }
    
    @SubscribeEvent
    public void gemSmashing(final AnvilLandEvent e) {
        final Level level = e.getLevel();
        final BlockPos pos = e.getPos();
        final List<ItemEntity> items = level.m_45976_((Class)ItemEntity.class, new AABB(pos, pos.m_7918_(1, 1, 1)));
        for (final ItemEntity ent : items) {
            final ItemStack stack = ent.m_32055_();
            if (stack.m_41720_() == Adventure.Items.GEM.get()) {
                ent.m_32045_(new ItemStack((ItemLike)Adventure.Items.GEM_DUST.get(), stack.m_41613_()));
            }
        }
    }
    
    @SubscribeEvent(priority = EventPriority.HIGH)
    public void enchLevels(final GetEnchantmentLevelEvent e) {
        final boolean isReentrant = AdventureEvents.reentrantLock.get().getAndSet(true);
        if (isReentrant) {
            return;
        }
        SocketHelper.getGems(e.getStack()).getEnchantmentLevels(e.getEnchantments());
        AffixHelper.streamAffixes(e.getStack()).forEach(inst -> inst.getEnchantmentLevels(e.getEnchantments()));
        AdventureEvents.reentrantLock.get().set(false);
    }
    
    @SubscribeEvent
    public void update(final LivingEvent.LivingTickEvent e) {
        final LivingEntity entity = e.getEntity();
        if (entity.getPersistentData().m_128441_("apoth.burns_in_sun") && entity.m_9236_().m_46461_() && !entity.m_9236_().f_46443_) {
            final float f = entity.m_213856_();
            final BlockPos blockpos = BlockPos.m_274561_(entity.m_20185_(), entity.m_20188_(), entity.m_20189_());
            final boolean flag = entity.m_20071_() || entity.f_146808_ || entity.f_146809_;
            if (f > 0.5f && entity.f_19796_.m_188501_() * 30.0f < (f - 0.4f) * 2.0f && !flag && entity.m_9236_().m_45527_(blockpos)) {
                entity.m_20254_(8);
            }
        }
    }
    
    @SubscribeEvent
    public void despawn(final MobSpawnEvent.AllowDespawn e) {
        final Mob entity = e.getEntity();
        if (entity instanceof final AbstractGolem g) {
            if (g.f_19797_ > 12000 && g.getPersistentData().m_128471_("apoth.boss")) {
                final Entity player = (Entity)g.m_9236_().m_45930_((Entity)g, -1.0);
                if (player != null) {
                    final double dist = player.m_20280_((Entity)g);
                    final int despawnDist = g.m_6095_().m_20674_().m_21611_();
                    final int dsDistSq = despawnDist * despawnDist;
                    if (dist > dsDistSq) {
                        e.setResult(Event.Result.ALLOW);
                    }
                }
            }
        }
    }
    
    @SubscribeEvent
    public void clone(final PlayerEvent.Clone e) {
        final int oldSeed = e.getOriginal().getPersistentData().m_128451_("apoth_reforge_seed");
        e.getEntity().getPersistentData().m_128405_("apoth_reforge_seed", oldSeed);
    }
    
    static {
        AdventureEvents.reentrantLock = ThreadLocal.withInitial(() -> new AtomicBoolean(false));
    }
}
